#!/usr/bin/env python
# -*- coding:utf-8 -*-


"""
Python script that showcases the encoding, solving and testing of an
ILP that groups a set of participants into multiple teams taking
into account their preferences for team position and partner.
It is equivalent to a multi-clique MEWC problem with predicate
constraints.

Copyright 2021 aferro (ORCID: 0000-0003-3830-3595)
"""


import random
from collections import defaultdict
#
import numpy as np
import cvxpy as cp
from cvxpy.settings import CPLEX
import networkx as nx


# ##############################################################################
# # SYNTHETIC DATASET CREATION
# ##############################################################################
N = 61
NUM_POSITIONS = 5
POS_ENCODING = 2 ** np.arange(NUM_POSITIONS)  # 1, 2, 4...
PREFERENCE_SCORES = np.array([1, 2/3, 1/3]) * (NUM_POSITIONS - 1)
MIN_NEIGHS_IF_NONZERO = 4  # at most N-1
IGNORED_PENALTY = 1 * (NUM_POSITIONS - 1)
K = max(2 * N, POS_ENCODING.sum())  # K needs to be bigger than any of these 2
# Create graph with random relationships between -1 and 1
g = nx.complete_graph(N)
for ori, dest in g.edges():
    r = (random.random() * 2 - 1) ** 3
    g.edges[ori, dest]["weight"] = r
g_edge_idxs = np.array(g.edges)  # shape (M, 2)

# Create random preferences for all users
prefs = np.zeros((N, NUM_POSITIONS))
p_idxs = np.array([random.sample(range(NUM_POSITIONS), len(PREFERENCE_SCORES))
                   for _ in range(N)])
for i, score in enumerate(PREFERENCE_SCORES):
    prefs[range(N), p_idxs[:, i]] = score


# ##############################################################################
# # VARIABLE PARTITION AND OPTIMIZATION OBJECTIVE
# ##############################################################################
# column partition in 3 sections: team IDs, indiv. vars, relationship vars
ind_vars = NUM_POSITIONS + 1  # one per position + x0
rel_vars = 7  # [rel_score, x<, x>, x=, z<, z>, xp]
num_rels = len(g.edges)
num_pvars, num_relvars = N*ind_vars, num_rels*rel_vars
beg1, beg2, beg3, num_vars = np.cumsum([0, N, num_pvars, num_relvars])
# indexes for column partitions
t_idxs = np.arange(beg1, beg2)
pos_idxs = np.array([np.arange(i, i+NUM_POSITIONS)
                     for i in np.arange(beg2, beg3, ind_vars)])
x0_idxs = np.arange(beg2, beg3, ind_vars) + NUM_POSITIONS
rel_idxs = np.arange(beg3, num_vars, rel_vars) + 0
xl_idxs = rel_idxs + 1
xg_idxs = rel_idxs + 2
xe_idxs = rel_idxs + 3
zl_idxs = rel_idxs + 4
zg_idxs = rel_idxs + 5
zp_idxs = rel_idxs + 6

# Create objective vector with 3 sections: team ID, preferences, relationships
c = np.zeros(num_vars)
for i, positions in enumerate(pos_idxs):
    c[positions] = prefs[i]
rel_weights = [g.edges[idx]["weight"] for idx in g_edge_idxs]
c[rel_idxs] = rel_weights
c[x0_idxs] = -IGNORED_PENALTY

# Create variables analogously to objective vector: teamID are int, rest bool
x_int = cp.Variable(N, integer=True)
x_bool = cp.Variable(num_pvars+num_relvars, boolean=True)
x = cp.hstack([x_int, x_bool])


# ##############################################################################
# # INEQUALITY CONSTRAINTS
# ##############################################################################
ineq_blocksizes = [N,  # goal 1
                   N, N,  # goal 2
                   num_rels, num_rels, num_rels, num_rels,  # goal 5
                   num_rels, num_rels, num_rels, num_rels,  # goal 6
                   num_rels, num_rels,  # goal 7
                   N, N]  # goal 8
ineq_begs = np.cumsum([0] + ineq_blocksizes)
total_ineq = ineq_begs[-1]
#
G = np.zeros((total_ineq, num_vars))
h = np.zeros(total_ineq)

# Goal 1: non-negativity of team IDs: -t <= 0
G[ineq_begs[0]:ineq_begs[1], t_idxs] = -np.eye(N)

# Goal 2: x_0 if and only if t=0 for each n: (-t - x_0 <= -1), (t + K*x_0 <= K)
G[ineq_begs[1]:ineq_begs[2], t_idxs] = -np.eye(N)
G[ineq_begs[1]:ineq_begs[2], x0_idxs] = -np.eye(N)
h[ineq_begs[1]:ineq_begs[2]] = -1
G[ineq_begs[2]:ineq_begs[3], t_idxs] = np.eye(N)
G[ineq_begs[2]:ineq_begs[3], x0_idxs] = np.eye(N) * K
h[ineq_begs[2]:ineq_begs[3]] = K

# Goal 5 ineq: teamID integer comparator for each relationship:
t_ori_idxs = t_idxs[g_edge_idxs[:, 0]]
t_dest_idxs = t_idxs[g_edge_idxs[:, 1]]
# xl constraints: (t1-t2) - K*xl <= 0      (t2-t1) + K*xl <= (K-1)
G[ineq_begs[3]:ineq_begs[4]][range(num_rels), t_ori_idxs] = 1
G[ineq_begs[3]:ineq_begs[4]][range(num_rels), t_dest_idxs] = -1
G[ineq_begs[3]:ineq_begs[4]][range(num_rels), xl_idxs] = -K
#
G[ineq_begs[4]:ineq_begs[5]][range(num_rels), t_ori_idxs] = -1
G[ineq_begs[4]:ineq_begs[5]][range(num_rels), t_dest_idxs] = 1
G[ineq_begs[4]:ineq_begs[5]][range(num_rels), xl_idxs] = K
h[ineq_begs[4]:ineq_begs[5]] = K - 1
# xg constraints: (t2-t1) - K*xg <= 0      (t1-t2) + K*xg <= (K-1)
G[ineq_begs[5]:ineq_begs[6]][range(num_rels), t_ori_idxs] = -1
G[ineq_begs[5]:ineq_begs[6]][range(num_rels), t_dest_idxs] = 1
G[ineq_begs[5]:ineq_begs[6]][range(num_rels), xg_idxs] = -K
#
G[ineq_begs[6]:ineq_begs[7]][range(num_rels), t_ori_idxs] = 1
G[ineq_begs[6]:ineq_begs[7]][range(num_rels), t_dest_idxs] = -1
G[ineq_begs[6]:ineq_begs[7]][range(num_rels), xg_idxs] = K
h[ineq_begs[6]:ineq_begs[7]] = K - 1

# Goal 6 ineq: binary-encoded position comparator for each relationship:
# zl constraint 1: (bin(pos1)-bin(pos2)) - K*zl <= 0
# zl constraint 2: (bin(pos2)-bin(pos1)) + K*zl <= (K-1)
for row_i, (ori, dest) in enumerate(g_edge_idxs):
    ori_positions, dest_positions = pos_idxs[ori], pos_idxs[dest]
    G[ineq_begs[7]+row_i][ori_positions] = POS_ENCODING
    G[ineq_begs[7]+row_i][dest_positions] = -POS_ENCODING
    G[ineq_begs[8]+row_i][ori_positions] = -POS_ENCODING
    G[ineq_begs[8]+row_i][dest_positions] = POS_ENCODING
G[ineq_begs[7]:ineq_begs[8]][range(num_rels), zl_idxs] = -K
G[ineq_begs[8]:ineq_begs[9]][range(num_rels), zl_idxs] = K
h[ineq_begs[8]:ineq_begs[9]] = K - 1
# zg constraint 1: (bin(pos2)-bin(pos1)) - K*zl <= 0
# zg constraint 2: (bin(pos1)-bin(pos2)) + K*zl <= (K-1)
for row_i, (ori, dest) in enumerate(g_edge_idxs):
    ori_positions, dest_positions = pos_idxs[ori], pos_idxs[dest]
    G[ineq_begs[9]+row_i][ori_positions] = -POS_ENCODING
    G[ineq_begs[9]+row_i][dest_positions] = POS_ENCODING
    G[ineq_begs[10]+row_i][ori_positions] = POS_ENCODING
    G[ineq_begs[10]+row_i][dest_positions] = -POS_ENCODING
G[ineq_begs[9]:ineq_begs[10]][range(num_rels), zg_idxs] = -K
G[ineq_begs[10]:ineq_begs[11]][range(num_rels), zg_idxs] = K
h[ineq_begs[10]:ineq_begs[11]] = K - 1

# Goal 7 ineq: feasibility and state of relationships based on team ID and pos
x0_ori_idxs = x0_idxs[g_edge_idxs[:, 0]]
x0_dest_idxs = x0_idxs[g_edge_idxs[:, 1]]
# Constraint 1 for each rel: (2xe-zp) - rel - K*x0 - K*x'0 <= 0
G[ineq_begs[11]:ineq_begs[12]][range(num_rels), xe_idxs] = 2
G[ineq_begs[11]:ineq_begs[12]][range(num_rels), zp_idxs] = -1
G[ineq_begs[11]:ineq_begs[12]][range(num_rels), rel_idxs] = -1
G[ineq_begs[11]:ineq_begs[12]][range(num_rels), x0_ori_idxs] = -K
G[ineq_begs[11]:ineq_begs[12]][range(num_rels), x0_dest_idxs] = -K
# Constraint 2 for each rel: K*rel -(2*xe-zp) - K*x0 - K*x'0 <= (K-1)
G[ineq_begs[12]:ineq_begs[13]][range(num_rels), rel_idxs] = K
G[ineq_begs[12]:ineq_begs[13]][range(num_rels), xe_idxs] = -2
G[ineq_begs[12]:ineq_begs[13]][range(num_rels), zp_idxs] = 1
G[ineq_begs[12]:ineq_begs[13]][range(num_rels), x0_ori_idxs] = -K
G[ineq_begs[12]:ineq_begs[13]][range(num_rels), x0_dest_idxs] = -K
h[ineq_begs[12]:ineq_begs[13]] = K - 1

# Goal 8: Enforce that each individual has either 0 relationships or >=NZ:
# K*x0 + sum(rels) <= K (if x0==1, sum(rels) must be zero)
G[ineq_begs[13]:ineq_begs[14]][range(N), x0_idxs] = K
for i in range(N):
    # For individual i, find var indexes of all its relationships
    i_edge_idxs = np.where((g_edge_idxs == i).any(axis=1))[0]
    all_i_rels = rel_idxs[i_edge_idxs]  # (N-1)
    # and set them
    G[ineq_begs[13]+i, all_i_rels] = 1
h[ineq_begs[13]:ineq_begs[14]] = K
# -NZ*x0 - sum(rels) <= -NZ  (if x0==0, sum(rels) must be >=NZ, otherwise DC)
G[ineq_begs[14]:ineq_begs[15]][range(N), x0_idxs] = -MIN_NEIGHS_IF_NONZERO
for i in range(N):
    # For individual i, find var indexes of all its relationships
    i_edge_idxs = np.where((g_edge_idxs == i).any(axis=1))[0]
    all_i_rels = rel_idxs[i_edge_idxs]  # (N-1)
    # and set them
    G[ineq_begs[14]+i, all_i_rels] = -1
h[ineq_begs[14]:ineq_begs[15]] = -MIN_NEIGHS_IF_NONZERO


# ##############################################################################
# # EQUALITY CONSTRAINTS
# ##############################################################################
eq_blocksizes = [N, N, num_rels, num_rels]
eq_begs = np.cumsum([0] + eq_blocksizes)
total_eq = sum(eq_blocksizes)
#
A = np.zeros((total_eq, num_vars))
b = np.zeros(total_eq)

# Goal 3: x_0 + sum(x_pos) = 1 for each n
for i, row_i in enumerate(range(eq_begs[0], eq_begs[1])):
    A[row_i, x0_idxs[i]] = 1
    A[row_i, pos_idxs[i]] = 1
b[eq_begs[0]:eq_begs[1]] = 1

# Goal 4: sum(discarded_positions) = 0 for each n
for i, row_i in enumerate(range(eq_begs[1], eq_begs[2])):
    discarded_by_i = (prefs[i] == 0)
    A[row_i, pos_idxs[i]] = discarded_by_i

# Goal 5 eq: enforce xe via xl+xg+xe = 1, once per relationship
A[eq_begs[2]:eq_begs[3]][range(num_rels), xl_idxs] = 1
A[eq_begs[2]:eq_begs[3]][range(num_rels), xg_idxs] = 1
A[eq_begs[2]:eq_begs[3]][range(num_rels), xe_idxs] = 1
b[eq_begs[2]:eq_begs[3]] = 1

# Goal 6 eq: enforce zp via zl+zg-zp = 0, once per relationship
A[eq_begs[3]:eq_begs[4]][range(num_rels), zl_idxs] = 1
A[eq_begs[3]:eq_begs[4]][range(num_rels), zg_idxs] = 1
A[eq_begs[3]:eq_begs[4]][range(num_rels), zp_idxs] = -1


breakpoint()

# ##############################################################################
# # SOLVER AND OPTIMIZATION
# ##############################################################################
prob = cp.Problem(cp.Maximize(c.T@x), [G@x <= h, A@x == b])
prob.solve(solver=CPLEX, verbose=True)  # CPLEX worked best with no tweaking


# ##############################################################################
# # TEST AND INSPECT RESULTS
# ##############################################################################
def iff(cond1, cond2):
    """
    :returns: True if both inputs have the same truth value, False otherwise.
      Inputs will be casted to ``bool``!
    """
    return not (bool(cond1) ^ bool(cond2))


# x has numeric noise that can affect bool tests. Round values to nearest int
x = np.around(x.value)

# collect result nodes and edges
result_nodes = {}
for i, (t_idx, x0_idx, pos_idx) in enumerate(zip(t_idxs, x0_idxs, pos_idxs)):
    result_nodes[i] = {"team_ID": round(x[t_idx]),
                       "pos_arr": x[pos_idx],
                       "pref_scores": c[pos_idx],
                       "x0": round(x[x0_idx])}
result_edges = {}
for i, (ori, dest) in enumerate(g_edge_idxs):
    result_edges[(ori, dest)] = {
        "weight": c[rel_idxs][i],
        "rel": x[rel_idxs][i],
        "xl": x[xl_idxs][i],
        "xg": x[xg_idxs][i],
        "xe": x[xe_idxs][i],
        "zl": x[zl_idxs][i],
        "zg": x[zg_idxs][i],
        "zp": x[zp_idxs][i]}

# gather results by team and node
nodes_by_team = defaultdict(dict)
for n, rn in result_nodes.items():
    nodes_by_team[rn["team_ID"]][n] = rn
edges_by_node = defaultdict(dict)
for (ori, dest), v in result_edges.items():
    edges_by_node[ori][(ori, dest)] = v
    edges_by_node[dest][(ori, dest)] = v
edges_by_team = defaultdict(dict)
for (ori, dest), v in result_edges.items():
    ori_n, dest_n = result_nodes[ori], result_nodes[dest]
    edges_by_team[ori_n["team_ID"]][(ori, dest)] = v
    edges_by_team[dest_n["team_ID"]][(ori, dest)] = v

# test that objectives were properly encoded
try:
    # Assert that x variable domains were respected:
    assert set(np.unique(x_bool.value.round())) == {0, 1}, \
        "x_bool var wasn't bool?"
    assert x.min() >= 0, "x variable had negative values!"

    # Test goal 1: all team IDs are nonnegative
    assert all(rn["team_ID"] >= 0
               for rn in result_nodes.values()), "Negative team ID given!"

    # Test goal 2: x0 is 1 if and only if team ID is 0
    assert all([iff(rn["x0"] == 1, rn["team_ID"] == 0)
                for rn in result_nodes.values()]), \
        "x0 if and only if teamID==0 was violated!"

    # Test goal 3: sum of x0 plus all position booleans is always 1
    assert all((rn["pos_arr"].sum() + rn["x0"]) == 1
               for rn in result_nodes.values()), \
        "x0+sum(position_booleans)==1 was violated!"

    # Test goal 4: all chosen position scores are >0, i.e. valid
    assert all(rn["pref_scores"].dot(rn["pos_arr"]) > 0
               for rn in result_nodes.values() if not rn["x0"]), \
        "Zero-score preference assigned to nonzero team!"

    # Test goal 5. For each pair: xe==1 IFF team1==team2. Also test xl and xg
    test5_data = [(result_nodes[ori]["team_ID"], result_nodes[dest]["team_ID"],
                   v["xl"], v["xg"], v["xe"])
                  for (ori, dest), v in result_edges.items()]
    assert all(iff((t2-t1) < 0, xl) for t1, t2, xl, _, _ in test5_data), \
        "xl does not correspond to t1 less than t2!"
    assert all(iff((t2-t1) > 0, xg) for t1, t2, _, xg, _ in test5_data), \
        "xg does not correspond to t1 greater than t2!"
    assert all(iff((t2-t1) == 0, xe) for t1, t2, _, _, xe in test5_data), \
        "xe does not correspond to t1 equals t2!"
    assert all((xl+xg+xe) == 1 for _, _, xl, xg, xe in test5_data), \
        "(xl+xg+xe)==1 did not hold!"

    # Test goal 6
    for (ori, dest), v in result_edges.items():
        ori_pos_code = POS_ENCODING.dot(result_nodes[ori]["pos_arr"])
        dest_pos_code = POS_ENCODING.dot(result_nodes[dest]["pos_arr"])
        pos_diff = dest_pos_code - ori_pos_code
        assert iff(pos_diff < 0, v["zl"]), "zl must correspond to pos1<pos2!"
        assert iff(pos_diff > 0, v["zg"]), "zg must correspond to pos1>pos2!"
        assert iff(pos_diff != 0, v["zp"]), "zp must correspond to pos1!=pos2!"

    # Test goal 7 for each relationship.
    # First test the consistency of the control variables
    for (ori, dest), v in result_edges.items():
        ori_n, dest_n = result_nodes[ori], result_nodes[dest]
        xe, zp, rel = v["xe"], v["zp"], v["rel"]
        # Whenever any x0==1, rel must be 0
        if (ori_n["x0"] + dest_n["x0"]) > 0:
            assert rel == 0, "Relationship involving zero-team was active!"
        else:  # if no team IDs are zero
            if xe == 0:  # if different teams, rel must be zero
                assert rel == 0, "Different nonzero teams had an active rel!"
            else:  # if same team:
                if zp == 1:  # ...and different positions, rel must be 1
                    assert rel == 1, \
                        "Same >0 team and diff pos, but rel wasn't active!"
                else:  # ...and same pos, should never exist!
                    raise AssertionError("xe==1 and zp==0 should never exist!")
    # Then test position consistency:
    for t, nodes in nodes_by_team.items():
        pos_codes = [POS_ENCODING.dot(n["pos_arr"]) for n in nodes.values()]
        if t == 0:
            # All zero-team positions are zeros
            assert all(code == 0 for code in pos_codes), \
                "0 team had nonzero pos!"
        else:
            # All nonzero teams have as many unique positions as members
            assert len(nodes) == len(set(pos_codes)), \
                "Nonzero team didn't have as many unique positions as members!"

    # Test goal 8 alt: all nodes have either 0 neighbours or >NZ
    for n, edges in edges_by_node.items():
        n_edges = len(edges)
        assert n_edges == 0 or n_edges >= MIN_NEIGHS_IF_NONZERO, \
            "All nodes must have either 0 or >={MIN_NEIGHS_IF_NONZERO} rels!"

    # Test goal 9: all teams are cliques, i.e:
    # Each node appears at most in 1 team
    # Each node in a K-clique has exactly (K-1) connections
    # (K-1) connections are between n and all the rest of clique members
    assert len(result_nodes) == N, "All nodes should have exactly 1 team!"
    for t, nodes in nodes_by_team.items():
        team_nodes = set(nodes.keys())
        t_size = len(nodes)
        for n in team_nodes:
            n_rels = [k for k, v in edges_by_node[n].items() if v["rel"]]
            assert len(n_rels) == t_size - 1, \
                "Node in a K-clique didn't have (K-1) rels!"
            assert all(n in rel for rel in n_rels), \
                "Connection for node n didn't contain n!"
            assert team_nodes == set(sum(n_rels, tuple())), \
                "(K-1) connections from node n didn't form the clique!"

    print("\n\nALL TESTS PASSED SUCCESSFULLY!")

except AssertionError as ae:
    print("ASSERTION ERROR (should never happen!):", ae)
    breakpoint()


# show some statistics for qualitative assessment
paired_n = [n for n, rn in result_nodes.items() if not rn["x0"]]
unpaired_n = [n for n, rn in result_nodes.items() if rn["x0"]]

pos_score_per_n = {n: rn["pref_scores"].dot(rn["pos_arr"])
                   for n, rn in result_nodes.items() if not rn["x0"]}
sum_pos_score = sum(pos_score_per_n.values())
active_pos_mean = sum_pos_score/len(pos_score_per_n) if sum_pos_score else 0
print("\n\nPOSITION PREFERENCE: domain/active_avg/all_avg/and per-node",
      PREFERENCE_SCORES, active_pos_mean, sum_pos_score/N,
      pos_score_per_n, sep="\n")

print("\n\nACTIVE RELATIONSHIPS PER TEAM (num_rels/min/mean/max):")
for t, edges in edges_by_team.items():
    t_rels = [v["weight"] for v in edges.values() if v["rel"]]
    if t_rels:
        print(t, len(t_rels), round(np.min(t_rels), 2),
              round(np.mean(t_rels), 2), round(np.max(t_rels), 2))
    else:
        print(t, len(t_rels))

print("\n\nINACTIVE RELATIONSHIPS PER TEAM (num_rels/min/mean/max):")
for t, edges in edges_by_team.items():
    t_rels = [v["weight"] for v in edges.values() if not v["rel"]]
    if t_rels:
        print(t, len(t_rels), round(np.min(t_rels), 2),
              round(np.mean(t_rels), 2), round(np.max(t_rels), 2))
    else:
        print(t, len(t_rels))

print("\n\nACTIVE RELATIONSHIPS PER NODE",
      "((team, pos)/num_rels/min/mean/max/list):")
for n, edges in edges_by_node.items():
    neigh_ids_ws = [(k, v["weight"]) for k, v in edges.items() if v["rel"]]
    n_team = result_nodes[n]["team_ID"]
    n_pos = POS_ENCODING.dot(result_nodes[n]["pos_arr"])
    print(" >", n, (n_team, n_pos), len(neigh_ids_ws))
    if neigh_ids_ws:
        neigh_ids, weights = zip(*neigh_ids_ws)
        print(round(np.min(weights), 2), round(np.mean(weights), 2),
              round(np.max(weights), 2), neigh_ids)

print("\n\nINACTIVE RELATIONSHIPS PER NODE",
      "((team, pos)/num_rels/min/mean/max/list):")
for n, edges in edges_by_node.items():
    neigh_ids_ws = [(k, v["weight"]) for k, v in edges.items() if not v["rel"]]
    n_team = result_nodes[n]["team_ID"]
    n_pos = POS_ENCODING.dot(result_nodes[n]["pos_arr"])
    print(" >", n, (n_team, n_pos), len(neigh_ids_ws))
    if neigh_ids_ws:
        neigh_ids, weights = zip(*neigh_ids_ws)
        print(round(np.min(weights), 2), round(np.mean(weights), 2),
              round(np.max(weights), 2), neigh_ids)

print("\n\n NODES PER TEAM: team, num_nodes, IDs")
for t, nodes in nodes_by_team.items():
    print(t, len(nodes), nodes.keys())

print("\n\nSOLVER (status, opt value):", prob.status, prob.value)


# useful plotting one-liners
import matplotlib.pyplot as plt
from matplotlib import colors
breakpoint()

# rng=(6, 7); plt.clf(); plt.imshow(np.hstack([G, h[:, np.newaxis]])[ineq_begs[rng[0]]:ineq_begs[rng[1]]], cmap="bwr", norm=colors.TwoSlopeNorm(vmin=-4, vcenter=0, vmax=4)); plt.show()
# rng=(0, 4); plt.clf(); plt.imshow(np.hstack([A, b[:, np.newaxis]])[eq_begs[rng[0]]:eq_begs[rng[1]]], cmap="bwr", norm=colors.TwoSlopeNorm(vmin=-4, vcenter=0, vmax=4)); plt.show()

# graph=g; edgelabels={k:v["weight"] for k, v in graph.edges.items()}; plt.clf(); pos=nx.spring_layout(graph); nx.draw(graph); nx.draw_networkx_edges(graph); plt.show()
# graph=g; plt.clf(); nx.draw_networkx(graph); pos=nx.spring_layout(graph); nx.draw_networkx_edges(graph); plt.show()

# rng=(0, 15); plt.clf(); plt.figure(figsize=(15, 15)); plt.imshow(np.hstack([G, h[:, np.newaxis]])[ineq_begs[rng[0]]:ineq_begs[rng[1]]], cmap="bwr", norm=colors.TwoSlopeNorm(vmin=-4, vcenter=0, vmax=4)); plt.savefig("inequalities.png", transparent=True, dpi=200)
# rng=(0, 4); plt.clf(); plt.figure(figsize=(15, 5)); plt.imshow(np.hstack([A, b[:, np.newaxis]])[ineq_begs[rng[0]]:ineq_begs[rng[1]]], cmap="bwr", norm=colors.TwoSlopeNorm(vmin=-4, vcenter=0, vmax=4)); plt.savefig("equalities.png", transparent=True, dpi=200)
